#include "PedMarkerData.h"

#include <cassert>
#include <string>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
#include <iostream>
#include "Argument.h"
#include "IO.h"
#include "tabix.h"
#include "PeopleSet.h"
#include "RangeList.h"
#include "VCFUtil.h"
#include "Error.h"

using namespace std;

PedMarkerData::PedMarkerData(){Initialize();}
PedMarkerData::~PedMarkerData()
{
  if(markers!=NULL) DeleteMarkerMemory();
  if(famSizes!=NULL) { delete [] famSizes; famSizes=NULL; }
}

PedMarkerData::PedMarkerData(Pedigree *p)
{
  Initialize();
  SetPedigree(p);
}

void PedMarkerData::Initialize()
{
  ped = NULL;
  markers = NULL;
  famSizes = NULL;
  size = INIT_SIZE;
  count = 0;
  nFam = 0;
  start = 0;
  end = 0;
  GL_idx = -1;
  PL_idx = -1;
  submap = -1.0;
  for (int i=0; i<=GLLIM; i++) PL2LK_table[i]=pow(10, -float(i)/10.0);
}

float PedMarkerData::PL2LK(int pl)
{
  if(pl<0) error("Phred-scaled likelihood %d can not be negative\n", pl);
  if(pl>GLLIM) return(PL2LK_table[GLLIM]);
  return(PL2LK_table[pl]);
}

void PedMarkerData::AllocateMarkerMemory()
{
  var.resize(size);
  
  markers = new SeqMarkerInfo **[size];
  for(int i=0; i<size; i++)
    {
      markers[i] = new SeqMarkerInfo *[nFam];
      
      for(int j=0; j<nFam; j++)
	markers[i][j] = new SeqMarkerInfo[famSizes[j]];
    }
}

void PedMarkerData::AllocateMarkerMemory(SeqMarkerInfo *** &new_markers)
{
  var.resize(size);
  
  SeqMarkerInfo ***markers = new SeqMarkerInfo **[size];
  
  for(int i=0; i<size; i++)
    {
      markers[i] = new SeqMarkerInfo*[nFam];
      
      for(int j=0; j<nFam; j++)
	markers[i][j] = new SeqMarkerInfo[famSizes[j]];
    }
  new_markers = markers;
}

void PedMarkerData::ReAllocateMarkerMemory()
{
  printf("Re-allocating marker memory...\n");
  size = size*2;
  
  var.resize(size);
  
  SeqMarkerInfo ***new_markers;
  AllocateMarkerMemory(new_markers);
  CopyMarkerMemory(new_markers);
  DeleteMarkerMemory();
  markers = new_markers;
}

void PedMarkerData::CopyMarkerMemory(SeqMarkerInfo ***new_markers)
{
  for(int i=0; i<count; i++)
    for(int j=0; j<nFam; j++)
      for(int k=0; k<famSizes[j]; k++)
	new_markers[i][j][k] = markers[i][j][k];
}

void PedMarkerData::DeleteMarkerMemory()
{
  if(markers==NULL) return;
  
  for(int i=0; i<size/2; i++) //size is floatd after re-allocation
    {
      for(int j=0; j<nFam; j++)
	delete [] markers[i][j];
      delete [] markers[i];
    }
  delete [] markers;
  markers = NULL;
}

void PedMarkerData::SetPedigree(Pedigree *p)
{
  ped = p;
  nFam = ped->familyCount;
  bool flag = false;

  for(int i=0; i<ped->familyCount; i++)
  {
   if(ped->families[i]->founders == ped->families[i]->count)
   {
     fprintf(stderr, "ERROR: Family %s has founders only. Please remove and rerun...\n", ped->families[i]->famid.c_str());
     flag=true;
   }
  }

  if(flag==true) exit(1);

  famSizes = new int[nFam];
  for(int i=0; i<nFam; i++)
    famSizes[i] = ped->families[i]->count;
  
  AllocateMarkerMemory();

  MapPID2Traverse();
}

void PedMarkerData::SetSubmap(float sub) { 
  submap = sub ;
}

void PedMarkerData::SetInputRange(std::string c, int s, int e)
{
  chr = c;
  start = s;
  end = e;
}

void PedMarkerData::MapPID2Traverse()
{
  pid2traverse.clear();
  pids.clear();
  traverse2pid.clear();

  for(int i=0; i<ped->familyCount; i++)
    for(int j=0; j<ped->families[i]->count; j++)
      {
	std::pair<int, int> traverse_pair;
	traverse_pair.first = i;
	traverse_pair.second = j;

	Person *p;
	p = ped->persons[ped->families[i]->path[j]];
	pid2traverse[std::string(p->pid.c_str())] = traverse_pair;
        traverse2pid[traverse_pair] = std::string(p->pid.c_str());
	pids.push_back(std::string(p->pid.c_str()));
      }
}

void PedMarkerData::LoadMarkerVCF(String vcf)
{
  printf("Loading VCF file for inferring IBD...\n");
  
  VCFInputFile vin(vcf.c_str());
  vin.includePeople(pids);

  GL_idx = -1;
  PL_idx = -1;
  count = 0;
  int map_counter = 0;

  while(vin.readRecord())
    {
      VCFRecord &r = vin.getVCFRecord();
      VCFPeople &people = r.getPeople();

      VCFIndividual* indv;

  std::string var_chr = r.getChrom();
  if(var_chr.substr(0,3).compare("chr")==0) var_chr = var_chr.substr(3, var_chr.size());

  if(pids_with_data.size()==0)
  for(int i=0; i<people.size(); i++)
  {
      indv = people[i];
      std::string pid;
      pid = (*indv).getName();

      if(pid2traverse.count(pid)==0) {printf("Sample -%s- is not included in the analysis\n", pid.c_str() ); continue; }

      std::pair<int, int> fam_pid_index_pair = pid2traverse[pid];

      fam_pid_index_pair_with_data[fam_pid_index_pair]++;

      pids_with_data.push_back(pid);
  }

  if(pids_with_data.size()==0) error("NO samples are in both PED and VCF files!\n");

      if(chr.size()>0)
        {
	  if(count==0 && chr!=var_chr ) continue;
	  if(chr!=var_chr ) break;
	  if(end>0)
	    if(r.getPos()<start) continue;
	    else if(r.getPos()>end) break;
        }
      
      if(r.getPos() != marker_map.pos[map_counter] ) continue;
      
      var[count].chr=var_chr;
      if(var[count].chr.substr(0,3).compare("chr")==0) var[count].chr = var[count].chr.substr(3, var[count].chr.size());
      var[count].pos = r.getPos();
      var[count].ref = r.getRef();
      var[count].alt = r.getAlt();
      
      bool missing = false;
      var[count].freq = 1-r.getInfoTag("AF", &missing).toDouble();
      if(var[count].freq>=1.0) var[count].freq = 0.999;
      if(var[count].freq<=0.0) var[count].freq = 0.001;

      if(missing) error("AF is not in the INFO in the VCF file! You can run POLYMUTT to calculate AF values\n");
      
      if(GL_idx<0 && PL_idx<0)
	{
	  GL_idx = r.getFormatIndex("GL");
	  PL_idx = r.getFormatIndex("PL");
	  
	  if(GL_idx<0 && PL_idx<0) {
	    fprintf(stderr, "NO GL or PL field was found. Please check the vcf file at chr:%s and position:%d", r.getChrom(), r.getPos());
	    exit(1);
	  }
	  
	}
      
      
      std::map<std::string, std::pair<int, int> >::iterator it;
      StringArray gl;
      //std::pair<int, int> idx_pair;

      for(int i=0; i<people.size(); i++)
        { 
	  indv = people[i];
	  if(pid2traverse.count((*indv).getName())==0) continue;
	  
	  //idx_pair = pid2traverse[(*indv).getName()];
	  it = pid2traverse.find( (*indv).getName() );
	  //if(it==pid2traverse.end() ) continue;

          bool missing = false;
          std::string pl_str = GL_idx>0 ? (*indv).get(GL_idx, &missing).toStr() : (*indv).get(PL_idx, &missing).toStr();

          if(missing || pl_str.compare(".")==0) {
            missing = true;
	  }
          else {
           if(pl_str.find(",")==string::npos) error("PL format seems wrong!\n");

           gl.ReplaceTokens(String(pl_str.c_str()), ",");

           if(gl.Length()<3) missing = true;
           else if(gl[0]=="." || gl[1]=="." || gl[2]==".") missing = true;
           else missing = false;
           }

          markers[count][it->second.first][it->second.second].missing_pl = missing;

	  if(missing) {
	    markers[count][it->second.first][it->second.second].geno_lk[0] = 1.0;
	    markers[count][it->second.first][it->second.second].geno_lk[1] = 1.0;
	    markers[count][it->second.first][it->second.second].geno_lk[2] = 1.0;
	    markers[count][it->second.first][it->second.second].geno_pl[0] = 0;
	    markers[count][it->second.first][it->second.second].geno_pl[1] = 0;
	    markers[count][it->second.first][it->second.second].geno_pl[2] = 0;
	  }
	  else{
	   float GL0 = gl[0].AsDouble();
	   float GL1 = gl[1].AsDouble();
	   float GL2 = gl[2].AsDouble();
	  
	   int PL0, PL1, PL2;
	   PL0 = int(PL_idx>0 ? GL0 : -10*GL0); if(PL0>GLLIM) PL0 = GLLIM;
	   PL1 = int(PL_idx>0 ? GL1 : -10*GL1); if(PL0>GLLIM) PL1 = GLLIM;
	   PL2 = int(PL_idx>0 ? GL2 : -10*GL2); if(PL0>GLLIM) PL2 = GLLIM;

	   markers[count][it->second.first][it->second.second].geno_lk[0] = PL2LK(PL0);
	   markers[count][it->second.first][it->second.second].geno_lk[1] = PL2LK(PL1);
	   markers[count][it->second.first][it->second.second].geno_lk[2] = PL2LK(PL2);
	   markers[count][it->second.first][it->second.second].geno_pl[0] = PL0;
	   markers[count][it->second.first][it->second.second].geno_pl[1] = PL1;
	   markers[count][it->second.first][it->second.second].geno_pl[2] = PL2;
	  }
        }
      count++;
      map_counter++;

      if(count%10000==0) printf("Loading %d variants...\n", count);
      if(map_counter>=marker_map.pos.size() ) break;
      if(count>=size) ReAllocateMarkerMemory();
    }

  //It is required that all marekrs in Mapfile have to be present in the VCF file
  if(map_counter!=marker_map.pos.size()) error("Not all markers in Mapfile were found in VCF: %d out of %d were found!\n", map_counter, marker_map.pos.size());
}

void PedMarkerData::LoadMapFile(String &file)
{
  
  FILE *fh = fopen(file.c_str(), "r");
  if(fh==NULL){
    fprintf(stderr, "Open map file failed!\n"); exit(1);}
  
  fprintf(stderr, "Loading map file ... ");
  
  LoadMapFile(fh);
  
  fclose(fh);
}

void PedMarkerData::LoadMapFile(FILE *fh)
{
  String s;
  StringArray tokens;
  
  marker_map.pos.clear();
  marker_map.recomb.clear();
  marker_map.marker_name.clear();
  marker_map.chr.clear();
  marker_map.distance.clear();
  
  int lines = 0;
  while(!feof(fh))
    {
      s.ReadLine(fh);
      tokens.ReplaceTokens(s);
      if(tokens.Length()==0) continue;
      
      lines++;
      if(lines==1) { printf("Skipping the header %s...\n", s.c_str()); continue; } //skip the header
      
      if(tokens.Length()<4) { fprintf(stderr, "Less than 4 columns are in the map file!\n"); exit(1); }
      std::string map_chr = std::string(tokens[0].c_str());
      if(map_chr.substr(0,3).compare("chr")==0) map_chr = map_chr.substr(3, map_chr.size());

      if(chr.size()>0)
        {
	  if(marker_map.pos.size()==0 && chr!=std::string(tokens[0].c_str()) ) continue;
	  if(chr.compare(map_chr)!=0) break;
	  if(end>0)
	    if(tokens[3].AsInteger()<start) continue;
	    else if(tokens[3].AsInteger()>end) break;
        }

      if(submap>0 && (rand()%100+1)>submap*100) continue;
      
      marker_map.chr.push_back(map_chr);
      marker_map.pos.push_back(tokens[3].AsInteger());
      marker_map.marker_name.push_back(std::string(tokens[1].c_str()));
      marker_map.distance.push_back(tokens[2].AsDouble());
      
      if(marker_map.pos.size()==1) continue;
      if(marker_map.distance[marker_map.distance.size()-1] < marker_map.distance[marker_map.distance.size()-2])
        {
	  fprintf(stderr, "Map not in order!\n"); exit(1);
        }
      
      // Here this par to the DistanceToRecombination() function is in Morgan and the input in Mapfile is in cMorgan
      marker_map.recomb.push_back(DistanceToRecombination((marker_map.distance[marker_map.distance.size()-1]-marker_map.distance[marker_map.distance.size()-2] )*0.01));
    }//end of reading map file

  printf("Total markers in the map file loaded: %d\n", marker_map.pos.size() );
  if(marker_map.pos.size()==0) error("Map info was not loaded! Check mapfile and vcf file for consistency!\n");
}

void PedMarkerData::PrintGL()
{
  for(int i=0; i<count; i++)
    {
      for(int j=0; j<nFam; j++)
	for(int k=0; k<famSizes[j]; k++)
	  printf("m=%d i=%d j=%d %f %f %f\n", i, j, k, markers[i][j][k].geno_lk[0],markers[i][j][k].geno_lk[1],markers[i][j][k].geno_lk[2]);
      printf("\n");
    }
}
